# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.Game do
  @moduledoc """
  This behaviour should be implemented by any Amethyst Game. It additionally
  contains functions that the internal connection handler code uses to more
  conveniently call a game's callbacks.
  """

  @doc """
  `instantiate/1` is called when a new instance of your game is created. You may start any additional
  processes you want under the `DynamicSupervisor` 'supervisor' and return a map with the PIDs to those
  processes, known as your references. Other callbacks in your game will receive these references and
  may use them to access and update state.
  """
  @callback instantiate(supervisor :: pid()) :: {:ok, state_refs :: map()} | {:error, reason :: term}
  @doc """
  `login/3` is called when a new player logs into a game instance.
  You may either :accept or :reject the player for whatever reason, avoid
  rejecting the player in your default game as that will disconnect the player.

  If you :accept the player, you must return a spawn position ({x, y, z}) and
  rotation ({yaw, pitch})

  Note that if no new players can join for any reason, your game should return false from `joinable?/1`.

  The PID received in 'from' will be the one that calls all callbacks
  caused directly by this player, such as their movement or interactions --
  You can expect that PID to never be used by any other player and that this same
  player will not use another PID until they disconnect from the server.

  - 'from' is the PID of the player's connection process.
  - 'player_cfg' is a keyword list containing the configuration passed by the game client
  - 'state_refs' are your references (see `instantiate/1`)
  """
  @callback login(from :: pid(), player_cfg :: keyword(), state_refs :: map()) :: {:accept, {x :: float(), y :: float(), z :: float()}, {yaw :: float(), pitch ::float()}} | :reject
  def login(%{:mod => mod, :refs => refs}, player_cfg) do
    mod.login(self(), player_cfg, refs)
  end
  @doc """
  `player_position/3` is called when a player moves. This function is called with the absolute coordinates
  that the player client expects. TODO: Teleport Player API.

  - 'from' is the PID of the player's connection process (see `login/3`).
  - 'x', 'y' and 'z' are the absolute coordinates that the player wants to move to.
  - `state_refs` are your references (see `instantiate/1`)
  """
  @callback player_position(from :: pid(), {x :: float(), y :: float(), z :: float()}, state_refs :: map()) :: :ok
  def player_position(%{:mod => mod, :refs => refs}, {x, y, z}) do
    mod.player_position(self(), {x, y, z}, refs)
  end
  @doc """
  `player_rotation/3` is called when a player rotates. This function is called with the absolute angles
  that the player client expects. TODO: Teleport Player API.

  - 'from' is the PID of the player's connection process (see `login/3`).
  - 'yaw' and 'pitch' are the angles the player expects to rotate to. These are in Minecraft's rotation format.
  - `state_refs` are your references (see `instantiate/1`)
  """
  @callback player_rotation(from :: pid(), {yaw :: float(), pitch :: float()}, state_refs :: map()) :: :ok
  def player_rotation(%{:mod => mod, :refs => refs}, {yaw, pitch}) do
    mod.player_rotation(self(), {yaw, pitch}, refs)
  end
  @doc """
  `accept_teleport/3` is called when a client accepts a teleportation as sent by the Synchronize Player Position
  packet (TODO: Teleport Player API). This lets you know that the client is now where you expect it to be.

  - 'from' is the PID of the player's connection process (see `login/3`).
  - 'id' is the teleport ID (TODO: Teleport Player API)
  - 'state_refs' are your references (see `instantiate/1`)
  """
  @callback accept_teleport(from :: pid(), id :: integer(), state_refs :: map()) :: :ok
  def accept_teleport(%{:mod => mod, :refs => refs}, id) do
    mod.accept_teleport(self(), id, refs)
  end
  @doc """
  `chat/3` is called when a player sends a chat message. Note that it is not how slash commands should be handled.

  - 'from' is the PID of the player's connection process (see `login/3`).
  - 'message' is the chat message sent by the player.
  - `state_refs` are your references (see `instantiate/1`)
  """
  @callback chat(from :: pid(), message :: binary(), state_refs :: map()) :: :ok
  def chat(%{:mod => mod, :refs => refs}, message) do
    mod.chat(self(), message, refs)
  end
  @doc """
  The terrain of a specific chunk column. This is automatically used to load chunks for a player.

  For now, this data must be formatted as a 3D list, indexed as [y][z][x].
  """
  @callback chunk(from :: pid(), {x :: integer(), z :: integer()}, state_refs :: map()) :: [[[{block_state :: pos_integer(), sky_light :: 0..15, block_light :: 0..15}]]]
  def chunk(%{:mod => mod, :refs => refs}, pos) do
    mod.chunk(self(), pos, refs)
  end
  @doc """
  Whether or not this game instance can be joined by a new player. This should include basic logic such as
  if joining makes sense, for instance if the game is full or if the game has already started.
  """
  @callback joinable?(state_refs :: map()) :: boolean()
  def joinable?(%{:mod => mod, :refs => refs}) do
    mod.joinable?(refs)
  end
end
